@@ -22,17 +22,31 @@
  *
  */
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include <errno.h>
 #include <stdint.h>
 #include <stdlib.h>
 
 #include <bluetooth/bluetooth.h>
-#include <bluetooth/uuid.h>
 
 #include <glib.h>
 
+#include "src/shared/util.h"
+#include "lib/uuid.h"
 #include "att.h"
 
+static inline void put_uuid_le(const bt_uuid_t *src, void *dst)
+{
+	if (src->type == BT_UUID16)
+		put_le16(src->value.u16, dst);
+	else
+		/* Convert from 128-bit BE to LE */
+		bswap_128(&src->value.u128, dst);
+}
+
 const char *att_ecode2str(uint8_t status)
 {
 	switch (status)  {
@@ -101,6 +115,9 @@ struct att_data_list *att_data_list_alloc(uint16_t num, uint16_t len)
 	struct att_data_list *list;
 	int i;
 
+	if (len > UINT8_MAX)
+		return NULL;
+
 	list = g_new0(struct att_data_list, 1);
 	list->len = len;
 	list->num = num;
@@ -113,38 +130,51 @@ struct att_data_list *att_data_list_alloc(uint16_t num, uint16_t len)
 	return list;
 }
 
-uint16_t enc_read_by_grp_req(uint16_t start, uint16_t end, bt_uuid_t *uuid,
-							uint8_t *pdu, int len)
+static void get_uuid(uint8_t type, const void *val, bt_uuid_t *uuid)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(start) + sizeof(end);
-	uint16_t length;
+	if (type == BT_UUID16)
+		bt_uuid16_create(uuid, get_le16(val));
+	else {
+		uint128_t u128;
+
+		/* Convert from 128-bit LE to BE */
+		bswap_128(val, &u128);
+		bt_uuid128_create(uuid, u128);
+	}
+}
+
+uint16_t enc_read_by_grp_req(uint16_t start, uint16_t end, bt_uuid_t *uuid,
+						uint8_t *pdu, size_t len)
+{
+	uint16_t uuid_len;
 
 	if (!uuid)
 		return 0;
 
 	if (uuid->type == BT_UUID16)
-		length = 2;
+		uuid_len = 2;
 	else if (uuid->type == BT_UUID128)
-		length = 16;
+		uuid_len = 16;
 	else
 		return 0;
 
-	if (len < min_len + length)
-		return 0;
-
+	/* Attribute Opcode (1 octet) */
 	pdu[0] = ATT_OP_READ_BY_GROUP_REQ;
-	att_put_u16(start, &pdu[1]);
-	att_put_u16(end, &pdu[3]);
+	/* Starting Handle (2 octets) */
+	put_le16(start, &pdu[1]);
+	/* Ending Handle (2 octets) */
+	put_le16(end, &pdu[3]);
+	/* Attribute Group Type (2 or 16 octet UUID) */
+	put_uuid_le(uuid, &pdu[5]);
 
-	att_put_uuid(*uuid, &pdu[5]);
-
-	return min_len + length;
+	return 5 + uuid_len;
 }
 
-uint16_t dec_read_by_grp_req(const uint8_t *pdu, int len, uint16_t *start,
+uint16_t dec_read_by_grp_req(const uint8_t *pdu, size_t len, uint16_t *start,
 						uint16_t *end, bt_uuid_t *uuid)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*start) + sizeof(*end);
+	const size_t min_len = sizeof(pdu[0]) + sizeof(*start) + sizeof(*end);
+	uint8_t type;
 
 	if (pdu == NULL)
 		return 0;
@@ -155,21 +185,23 @@ uint16_t dec_read_by_grp_req(const uint8_t *pdu, int len, uint16_t *start,
 	if (pdu[0] != ATT_OP_READ_BY_GROUP_REQ)
 		return 0;
 
-	if (len < min_len + 2)
+	if (len == (min_len + 2))
+		type = BT_UUID16;
+	else if (len == (min_len + 16))
+		type = BT_UUID128;
+	else
 		return 0;
 
-	*start = att_get_u16(&pdu[1]);
-	*end = att_get_u16(&pdu[3]);
-	if (len == min_len + 2)
-		*uuid = att_get_uuid16(&pdu[5]);
-	else
-		*uuid = att_get_uuid128(&pdu[5]);
+	*start = get_le16(&pdu[1]);
+	*end = get_le16(&pdu[3]);
+
+	get_uuid(type, &pdu[5], uuid);
 
 	return len;
 }
 
 uint16_t enc_read_by_grp_resp(struct att_data_list *list, uint8_t *pdu,
-								int len)
+								size_t len)
 {
 	int i;
 	uint16_t w;
@@ -178,7 +210,7 @@ uint16_t enc_read_by_grp_resp(struct att_data_list *list, uint8_t *pdu,
 	if (list == NULL)
 		return 0;
 
-	if (len < list->len + 2)
+	if (len < list->len + sizeof(uint8_t) * 2)
 		return 0;
 
 	pdu[0] = ATT_OP_READ_BY_GROUP_RESP;
@@ -195,7 +227,7 @@ uint16_t enc_read_by_grp_resp(struct att_data_list *list, uint8_t *pdu,
 	return w;
 }
 
-struct att_data_list *dec_read_by_grp_resp(const uint8_t *pdu, int len)
+struct att_data_list *dec_read_by_grp_resp(const uint8_t *pdu, size_t len)
 {
 	struct att_data_list *list;
 	const uint8_t *ptr;
@@ -205,9 +237,29 @@ struct att_data_list *dec_read_by_grp_resp(const uint8_t *pdu, int len)
 	if (pdu[0] != ATT_OP_READ_BY_GROUP_RESP)
 		return NULL;
 
+	/* PDU must contain at least:
+	 * - Attribute Opcode (1 octet)
+	 * - Length (1 octet)
+	 * - Attribute Data List (at least one entry):
+	 *   - Attribute Handle (2 octets)
+	 *   - End Group Handle (2 octets)
+	 *   - Attribute Value (at least 1 octet) */
+	if (len < 7)
+		return NULL;
+
 	elen = pdu[1];
+	/* Minimum Attribute Data List size */
+	if (elen < 5)
+		return NULL;
+
+	/* Reject incomplete Attribute Data List */
+	if ((len - 2) % elen)
+		return NULL;
+
 	num = (len - 2) / elen;
 	list = att_data_list_alloc(num, elen);
+	if (list == NULL)
+		return NULL;
 
 	ptr = &pdu[2];
 
@@ -220,7 +272,8 @@ struct att_data_list *dec_read_by_grp_resp(const uint8_t *pdu, int len)
 }
 
 uint16_t enc_find_by_type_req(uint16_t start, uint16_t end, bt_uuid_t *uuid,
-			const uint8_t *value, int vlen, uint8_t *pdu, int len)
+					const uint8_t *value, size_t vlen,
+					uint8_t *pdu, size_t len)
 {
 	uint16_t min_len = sizeof(pdu[0]) + sizeof(start) + sizeof(end) +
 							sizeof(uint16_t);
@@ -234,16 +287,13 @@ uint16_t enc_find_by_type_req(uint16_t start, uint16_t end, bt_uuid_t *uuid,
 	if (uuid->type != BT_UUID16)
 		return 0;
 
-	if (len < min_len)
-		return 0;
-
 	if (vlen > len - min_len)
 		vlen = len - min_len;
 
 	pdu[0] = ATT_OP_FIND_BY_TYPE_REQ;
-	att_put_u16(start, &pdu[1]);
-	att_put_u16(end, &pdu[3]);
-	att_put_uuid16(*uuid, &pdu[5]);
+	put_le16(start, &pdu[1]);
+	put_le16(end, &pdu[3]);
+	put_le16(uuid->value.u16, &pdu[5]);
 
 	if (vlen > 0) {
 		memcpy(&pdu[7], value, vlen);
@@ -253,83 +303,84 @@ uint16_t enc_find_by_type_req(uint16_t start, uint16_t end, bt_uuid_t *uuid,
 	return min_len;
 }
 
-uint16_t dec_find_by_type_req(const uint8_t *pdu, int len, uint16_t *start,
-		uint16_t *end, bt_uuid_t *uuid, uint8_t *value, int *vlen)
+uint16_t dec_find_by_type_req(const uint8_t *pdu, size_t len, uint16_t *start,
+						uint16_t *end, bt_uuid_t *uuid,
+						uint8_t *value, size_t *vlen)
 {
-	int valuelen;
-	uint16_t min_len = sizeof(pdu[0]) + sizeof(*start) +
-						sizeof(*end) + sizeof(uint16_t);
-
 	if (pdu == NULL)
 		return 0;
 
-	if (len < min_len)
+	if (len < 7)
 		return 0;
 
+	/* Attribute Opcode (1 octet) */
 	if (pdu[0] != ATT_OP_FIND_BY_TYPE_REQ)
 		return 0;
 
-	/* First requested handle number */
-	if (start)
-		*start = att_get_u16(&pdu[1]);
-
-	/* Last requested handle number */
-	if (end)
-		*end = att_get_u16(&pdu[3]);
-
-	/* Always UUID16 */
-	if (uuid)
-		*uuid = att_get_uuid16(&pdu[5]);
-
-	valuelen = len - min_len;
+	/* First requested handle number (2 octets) */
+	*start = get_le16(&pdu[1]);
+	/* Last requested handle number (2 octets) */
+	*end = get_le16(&pdu[3]);
+	/* 16-bit UUID to find (2 octets) */
+	bt_uuid16_create(uuid, get_le16(&pdu[5]));
 
 	/* Attribute value to find */
-	if (valuelen > 0 && value)
-		memcpy(value, pdu + min_len, valuelen);
-
-	if (vlen)
-		*vlen = valuelen;
+	*vlen = len - 7;
+	if (*vlen > 0)
+		memcpy(value, pdu + 7, *vlen);
 
 	return len;
 }
 
-uint16_t enc_find_by_type_resp(GSList *matches, uint8_t *pdu, int len)
+uint16_t enc_find_by_type_resp(GSList *matches, uint8_t *pdu, size_t len)
 {
 	GSList *l;
 	uint16_t offset;
 
-	if (pdu == NULL || len < 5)
+	if (!pdu)
 		return 0;
 
 	pdu[0] = ATT_OP_FIND_BY_TYPE_RESP;
 
-	for (l = matches, offset = 1; l && len >= (offset + 4);
-					l = l->next, offset += 4) {
+	for (l = matches, offset = 1;
+				l && len >= (offset + sizeof(uint16_t) * 2);
+				l = l->next, offset += sizeof(uint16_t) * 2) {
 		struct att_range *range = l->data;
 
-		att_put_u16(range->start, &pdu[offset]);
-		att_put_u16(range->end, &pdu[offset + 2]);
+		put_le16(range->start, &pdu[offset]);
+		put_le16(range->end, &pdu[offset + 2]);
 	}
 
 	return offset;
 }
 
-GSList *dec_find_by_type_resp(const uint8_t *pdu, int len)
+GSList *dec_find_by_type_resp(const uint8_t *pdu, size_t len)
 {
 	struct att_range *range;
 	GSList *matches;
-	int offset;
+	off_t offset;
 
+	/* PDU should contain at least:
+	 * - Attribute Opcode (1 octet)
+	 * - Handles Information List (at least one entry):
+	 *   - Found Attribute Handle (2 octets)
+	 *   - Group End Handle (2 octets) */
 	if (pdu == NULL || len < 5)
 		return NULL;
 
 	if (pdu[0] != ATT_OP_FIND_BY_TYPE_RESP)
 		return NULL;
 
-	for (offset = 1, matches = NULL; len >= (offset + 4); offset += 4) {
+	/* Reject incomplete Handles Information List */
+	if ((len - 1) % 4)
+		return NULL;
+
+	for (offset = 1, matches = NULL;
+				len >= (offset + sizeof(uint16_t) * 2);
+				offset += sizeof(uint16_t) * 2) {
 		range = g_new0(struct att_range, 1);
-		range->start = att_get_u16(&pdu[offset]);
-		range->end = att_get_u16(&pdu[offset + 2]);
+		range->start = get_le16(&pdu[offset]);
+		range->end = get_le16(&pdu[offset + 2]);
 
 		matches = g_slist_append(matches, range);
 	}
@@ -338,37 +389,37 @@ GSList *dec_find_by_type_resp(const uint8_t *pdu, int len)
 }
 
 uint16_t enc_read_by_type_req(uint16_t start, uint16_t end, bt_uuid_t *uuid,
-							uint8_t *pdu, int len)
+						uint8_t *pdu, size_t len)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(start) + sizeof(end);
-	uint16_t length;
+	uint16_t uuid_len;
 
 	if (!uuid)
 		return 0;
 
 	if (uuid->type == BT_UUID16)
-		length = 2;
+		uuid_len = 2;
 	else if (uuid->type == BT_UUID128)
-		length = 16;
+		uuid_len = 16;
 	else
 		return 0;
 
-	if (len < min_len + length)
-		return 0;
-
+	/* Attribute Opcode (1 octet) */
 	pdu[0] = ATT_OP_READ_BY_TYPE_REQ;
-	att_put_u16(start, &pdu[1]);
-	att_put_u16(end, &pdu[3]);
+	/* Starting Handle (2 octets) */
+	put_le16(start, &pdu[1]);
+	/* Ending Handle (2 octets) */
+	put_le16(end, &pdu[3]);
+	/* Attribute Type (2 or 16 octet UUID) */
+	put_uuid_le(uuid, &pdu[5]);
 
-	att_put_uuid(*uuid, &pdu[5]);
-
-	return min_len + length;
+	return 5 + uuid_len;
 }
 
-uint16_t dec_read_by_type_req(const uint8_t *pdu, int len, uint16_t *start,
+uint16_t dec_read_by_type_req(const uint8_t *pdu, size_t len, uint16_t *start,
 						uint16_t *end, bt_uuid_t *uuid)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*start) + sizeof(*end);
+	const size_t min_len = sizeof(pdu[0]) + sizeof(*start) + sizeof(*end);
+	uint8_t type;
 
 	if (pdu == NULL)
 		return 0;
@@ -376,27 +427,29 @@ uint16_t dec_read_by_type_req(const uint8_t *pdu, int len, uint16_t *start,
 	if (start == NULL || end == NULL || uuid == NULL)
 		return 0;
 
-	if (len < min_len + 2)
+	if (len == (min_len + 2))
+		type = BT_UUID16;
+	else if (len == (min_len + 16))
+		type = BT_UUID128;
+	else
 		return 0;
 
 	if (pdu[0] != ATT_OP_READ_BY_TYPE_REQ)
 		return 0;
 
-	*start = att_get_u16(&pdu[1]);
-	*end = att_get_u16(&pdu[3]);
+	*start = get_le16(&pdu[1]);
+	*end = get_le16(&pdu[3]);
 
-	if (len == min_len + 2)
-		*uuid = att_get_uuid16(&pdu[5]);
-	else
-		*uuid = att_get_uuid128(&pdu[5]);
+	get_uuid(type, &pdu[5], uuid);
 
 	return len;
 }
 
-uint16_t enc_read_by_type_resp(struct att_data_list *list, uint8_t *pdu, int len)
+uint16_t enc_read_by_type_resp(struct att_data_list *list, uint8_t *pdu,
+								size_t len)
 {
 	uint8_t *ptr;
-	int i, w, l;
+	size_t i, w, l;
 
 	if (list == NULL)
 		return 0;
@@ -419,7 +472,7 @@ uint16_t enc_read_by_type_resp(struct att_data_list *list, uint8_t *pdu, int len
 	return w;
 }
 
-struct att_data_list *dec_read_by_type_resp(const uint8_t *pdu, int len)
+struct att_data_list *dec_read_by_type_resp(const uint8_t *pdu, size_t len)
 {
 	struct att_data_list *list;
 	const uint8_t *ptr;
@@ -429,9 +482,28 @@ struct att_data_list *dec_read_by_type_resp(const uint8_t *pdu, int len)
 	if (pdu[0] != ATT_OP_READ_BY_TYPE_RESP)
 		return NULL;
 
+	/* PDU must contain at least:
+	 * - Attribute Opcode (1 octet)
+	 * - Length (1 octet)
+	 * - Attribute Data List (at least one entry):
+	 *   - Attribute Handle (2 octets)
+	 *   - Attribute Value (at least 1 octet) */
+	if (len < 5)
+		return NULL;
+
 	elen = pdu[1];
+	/* Minimum Attribute Data List size */
+	if (elen < 3)
+		return NULL;
+
+	/* Reject incomplete Attribute Data List */
+	if ((len - 2) % elen)
+		return NULL;
+
 	num = (len - 2) / elen;
 	list = att_data_list_alloc(num, elen);
+	if (list == NULL)
+		return NULL;
 
 	ptr = &pdu[2];
 
@@ -443,22 +515,19 @@ struct att_data_list *dec_read_by_type_resp(const uint8_t *pdu, int len)
 	return list;
 }
 
-uint16_t enc_write_cmd(uint16_t handle, const uint8_t *value, int vlen,
-							uint8_t *pdu, int len)
+uint16_t enc_write_cmd(uint16_t handle, const uint8_t *value, size_t vlen,
+						uint8_t *pdu, size_t len)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(handle);
 
 	if (pdu == NULL)
 		return 0;
 
-	if (len < min_len)
-		return 0;
-
 	if (vlen > len - min_len)
 		vlen = len - min_len;
 
 	pdu[0] = ATT_OP_WRITE_CMD;
-	att_put_u16(handle, &pdu[1]);
+	put_le16(handle, &pdu[1]);
 
 	if (vlen > 0) {
 		memcpy(&pdu[3], value, vlen);
@@ -468,8 +537,8 @@ uint16_t enc_write_cmd(uint16_t handle, const uint8_t *value, int vlen,
 	return min_len;
 }
 
-uint16_t dec_write_cmd(const uint8_t *pdu, int len, uint16_t *handle,
-						uint8_t *value, int *vlen)
+uint16_t dec_write_cmd(const uint8_t *pdu, size_t len, uint16_t *handle,
+						uint8_t *value, size_t *vlen)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*handle);
 
@@ -485,29 +554,26 @@ uint16_t dec_write_cmd(const uint8_t *pdu, int len, uint16_t *handle,
 	if (pdu[0] != ATT_OP_WRITE_CMD)
 		return 0;
 
-	*handle = att_get_u16(&pdu[1]);
+	*handle = get_le16(&pdu[1]);
 	memcpy(value, pdu + min_len, len - min_len);
 	*vlen = len - min_len;
 
 	return len;
 }
 
-uint16_t enc_write_req(uint16_t handle, const uint8_t *value, int vlen,
-							uint8_t *pdu, int len)
+uint16_t enc_write_req(uint16_t handle, const uint8_t *value, size_t vlen,
+						uint8_t *pdu, size_t len)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(handle);
 
 	if (pdu == NULL)
 		return 0;
 
-	if (len < min_len)
-		return 0;
-
 	if (vlen > len - min_len)
 		vlen = len - min_len;
 
 	pdu[0] = ATT_OP_WRITE_REQ;
-	att_put_u16(handle, &pdu[1]);
+	put_le16(handle, &pdu[1]);
 
 	if (vlen > 0) {
 		memcpy(&pdu[3], value, vlen);
@@ -517,8 +583,8 @@ uint16_t enc_write_req(uint16_t handle, const uint8_t *value, int vlen,
 	return min_len;
 }
 
-uint16_t dec_write_req(const uint8_t *pdu, int len, uint16_t *handle,
-						uint8_t *value, int *vlen)
+uint16_t dec_write_req(const uint8_t *pdu, size_t len, uint16_t *handle,
+						uint8_t *value, size_t *vlen)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*handle);
 
@@ -534,7 +600,7 @@ uint16_t dec_write_req(const uint8_t *pdu, int len, uint16_t *handle,
 	if (pdu[0] != ATT_OP_WRITE_REQ)
 		return 0;
 
-	*handle = att_get_u16(&pdu[1]);
+	*handle = get_le16(&pdu[1]);
 	*vlen = len - min_len;
 	if (*vlen > 0)
 		memcpy(value, pdu + min_len, *vlen);
@@ -542,7 +608,7 @@ uint16_t dec_write_req(const uint8_t *pdu, int len, uint16_t *handle,
 	return len;
 }
 
-uint16_t enc_write_resp(uint8_t *pdu, int len)
+uint16_t enc_write_resp(uint8_t *pdu)
 {
 	if (pdu == NULL)
 		return 0;
@@ -552,7 +618,7 @@ uint16_t enc_write_resp(uint8_t *pdu, int len)
 	return sizeof(pdu[0]);
 }
 
-uint16_t dec_write_resp(const uint8_t *pdu, int len)
+uint16_t dec_write_resp(const uint8_t *pdu, size_t len)
 {
 	if (pdu == NULL)
 		return 0;
@@ -563,42 +629,36 @@ uint16_t dec_write_resp(const uint8_t *pdu, int len)
 	return len;
 }
 
-uint16_t enc_read_req(uint16_t handle, uint8_t *pdu, int len)
+uint16_t enc_read_req(uint16_t handle, uint8_t *pdu, size_t len)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(handle);
-
 	if (pdu == NULL)
 		return 0;
 
-	if (len < min_len)
-		return 0;
-
+	/* Attribute Opcode (1 octet) */
 	pdu[0] = ATT_OP_READ_REQ;
-	att_put_u16(handle, &pdu[1]);
+	/* Attribute Handle (2 octets) */
+	put_le16(handle, &pdu[1]);
 
-	return min_len;
+	return 3;
 }
 
 uint16_t enc_read_blob_req(uint16_t handle, uint16_t offset, uint8_t *pdu,
-									int len)
+								size_t len)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(handle) +
-							sizeof(offset);
-
 	if (pdu == NULL)
 		return 0;
 
-	if (len < min_len)
-		return 0;
-
+	/* Attribute Opcode (1 octet) */
 	pdu[0] = ATT_OP_READ_BLOB_REQ;
-	att_put_u16(handle, &pdu[1]);
-	att_put_u16(offset, &pdu[3]);
+	/* Attribute Handle (2 octets) */
+	put_le16(handle, &pdu[1]);
+	/* Value Offset (2 octets) */
+	put_le16(offset, &pdu[3]);
 
-	return min_len;
+	return 5;
 }
 
-uint16_t dec_read_req(const uint8_t *pdu, int len, uint16_t *handle)
+uint16_t dec_read_req(const uint8_t *pdu, size_t len, uint16_t *handle)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*handle);
 
@@ -614,12 +674,12 @@ uint16_t dec_read_req(const uint8_t *pdu, int len, uint16_t *handle)
 	if (pdu[0] != ATT_OP_READ_REQ)
 		return 0;
 
-	*handle = att_get_u16(&pdu[1]);
+	*handle = get_le16(&pdu[1]);
 
 	return min_len;
 }
 
-uint16_t dec_read_blob_req(const uint8_t *pdu, int len, uint16_t *handle,
+uint16_t dec_read_blob_req(const uint8_t *pdu, size_t len, uint16_t *handle,
 							uint16_t *offset)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*handle) +
@@ -640,13 +700,13 @@ uint16_t dec_read_blob_req(const uint8_t *pdu, int len, uint16_t *handle,
 	if (pdu[0] != ATT_OP_READ_BLOB_REQ)
 		return 0;
 
-	*handle = att_get_u16(&pdu[1]);
-	*offset = att_get_u16(&pdu[3]);
+	*handle = get_le16(&pdu[1]);
+	*offset = get_le16(&pdu[3]);
 
 	return min_len;
 }
 
-uint16_t enc_read_resp(uint8_t *value, int vlen, uint8_t *pdu, int len)
+uint16_t enc_read_resp(uint8_t *value, size_t vlen, uint8_t *pdu, size_t len)
 {
 	if (pdu == NULL)
 		return 0;
@@ -664,8 +724,8 @@ uint16_t enc_read_resp(uint8_t *value, int vlen, uint8_t *pdu, int len)
 	return vlen + 1;
 }
 
-uint16_t enc_read_blob_resp(uint8_t *value, int vlen, uint16_t offset,
-							uint8_t *pdu, int len)
+uint16_t enc_read_blob_resp(uint8_t *value, size_t vlen, uint16_t offset,
+						uint8_t *pdu, size_t len)
 {
 	if (pdu == NULL)
 		return 0;
@@ -681,61 +741,58 @@ uint16_t enc_read_blob_resp(uint8_t *value, int vlen, uint16_t offset,
 	return vlen + 1;
 }
 
-uint16_t dec_read_resp(const uint8_t *pdu, int len, uint8_t *value, int *vlen)
+ssize_t dec_read_resp(const uint8_t *pdu, size_t len, uint8_t *value,
+								size_t vlen)
 {
 	if (pdu == NULL)
-		return 0;
-
-	if (value == NULL || vlen == NULL)
-		return 0;
+		return -EINVAL;
 
 	if (pdu[0] != ATT_OP_READ_RESP)
-		return 0;
+		return -EINVAL;
+
+	if (value == NULL)
+		return len - 1;
+
+	if (vlen < (len - 1))
+		return -ENOBUFS;
 
 	memcpy(value, pdu + 1, len - 1);
 
-	*vlen = len - 1;
-
-	return len;
+	return len - 1;
 }
 
 uint16_t enc_error_resp(uint8_t opcode, uint16_t handle, uint8_t status,
-							uint8_t *pdu, int len)
+						uint8_t *pdu, size_t len)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(opcode) +
-						sizeof(handle) + sizeof(status);
-	uint16_t u16;
-
-	if (len < min_len)
-		return 0;
-
-	u16 = htobs(handle);
+	/* Attribute Opcode (1 octet) */
 	pdu[0] = ATT_OP_ERROR;
+	/* Request Opcode In Error (1 octet) */
 	pdu[1] = opcode;
-	memcpy(&pdu[2], &u16, sizeof(u16));
+	/* Attribute Handle In Error (2 octets) */
+	put_le16(handle, &pdu[2]);
+	/* Error Code (1 octet) */
 	pdu[4] = status;
 
-	return min_len;
+	return 5;
 }
 
-uint16_t enc_find_info_req(uint16_t start, uint16_t end, uint8_t *pdu, int len)
+uint16_t enc_find_info_req(uint16_t start, uint16_t end, uint8_t *pdu,
+								size_t len)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(start) + sizeof(end);
-
 	if (pdu == NULL)
 		return 0;
 
-	if (len < min_len)
-		return 0;
-
+	/* Attribute Opcode (1 octet) */
 	pdu[0] = ATT_OP_FIND_INFO_REQ;
-	att_put_u16(start, &pdu[1]);
-	att_put_u16(end, &pdu[3]);
+	/* Starting Handle (2 octets) */
+	put_le16(start, &pdu[1]);
+	/* Ending Handle (2 octets) */
+	put_le16(end, &pdu[3]);
 
-	return min_len;
+	return 5;
 }
 
-uint16_t dec_find_info_req(const uint8_t *pdu, int len, uint16_t *start,
+uint16_t dec_find_info_req(const uint8_t *pdu, size_t len, uint16_t *start,
 								uint16_t *end)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*start) + sizeof(*end);
@@ -752,17 +809,17 @@ uint16_t dec_find_info_req(const uint8_t *pdu, int len, uint16_t *start,
 	if (pdu[0] != ATT_OP_FIND_INFO_REQ)
 		return 0;
 
-	*start = att_get_u16(&pdu[1]);
-	*end = att_get_u16(&pdu[3]);
+	*start = get_le16(&pdu[1]);
+	*end = get_le16(&pdu[3]);
 
 	return min_len;
 }
 
 uint16_t enc_find_info_resp(uint8_t format, struct att_data_list *list,
-							uint8_t *pdu, int len)
+						uint8_t *pdu, size_t len)
 {
 	uint8_t *ptr;
-	int i, w;
+	size_t i, w;
 
 	if (pdu == NULL)
 		return 0;
@@ -770,7 +827,7 @@ uint16_t enc_find_info_resp(uint8_t format, struct att_data_list *list,
 	if (list == NULL)
 		return 0;
 
-	if (len < list->len + 2)
+	if (len < list->len + sizeof(uint8_t) * 2)
 		return 0;
 
 	pdu[0] = ATT_OP_FIND_INFO_RESP;
@@ -786,7 +843,7 @@ uint16_t enc_find_info_resp(uint8_t format, struct att_data_list *list,
 	return w;
 }
 
-struct att_data_list *dec_find_info_resp(const uint8_t *pdu, int len,
+struct att_data_list *dec_find_info_resp(const uint8_t *pdu, size_t len,
 							uint8_t *format)
 {
 	struct att_data_list *list;
@@ -815,6 +872,8 @@ struct att_data_list *dec_find_info_resp(const uint8_t *pdu, int len,
 	ptr = (void *) &pdu[2];
 
 	list = att_data_list_alloc(num, elen);
+	if (list == NULL)
+		return NULL;
 
 	for (i = 0; i < num; i++) {
 		memcpy(list->data[i], ptr, list->len);
@@ -824,8 +883,8 @@ struct att_data_list *dec_find_info_resp(const uint8_t *pdu, int len,
 	return list;
 }
 
-uint16_t enc_notification(uint16_t handle, uint8_t *value, int vlen,
-						uint8_t *pdu, int len)
+uint16_t enc_notification(uint16_t handle, uint8_t *value, size_t vlen,
+						uint8_t *pdu, size_t len)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(uint16_t);
 
@@ -836,14 +895,14 @@ uint16_t enc_notification(uint16_t handle, uint8_t *value, int vlen,
 		return 0;
 
 	pdu[0] = ATT_OP_HANDLE_NOTIFY;
-	att_put_u16(handle, &pdu[1]);
+	put_le16(handle, &pdu[1]);
 	memcpy(&pdu[3], value, vlen);
 
 	return vlen + min_len;
 }
 
-uint16_t enc_indication(uint16_t handle, uint8_t *value, int vlen,
-						uint8_t *pdu, int len)
+uint16_t enc_indication(uint16_t handle, uint8_t *value, size_t vlen,
+						uint8_t *pdu, size_t len)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(uint16_t);
 
@@ -854,14 +913,14 @@ uint16_t enc_indication(uint16_t handle, uint8_t *value, int vlen,
 		return 0;
 
 	pdu[0] = ATT_OP_HANDLE_IND;
-	att_put_u16(handle, &pdu[1]);
+	put_le16(handle, &pdu[1]);
 	memcpy(&pdu[3], value, vlen);
 
 	return vlen + min_len;
 }
 
-uint16_t dec_indication(const uint8_t *pdu, int len, uint16_t *handle,
-						uint8_t *value, int vlen)
+uint16_t dec_indication(const uint8_t *pdu, size_t len, uint16_t *handle,
+						uint8_t *value, size_t vlen)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(uint16_t);
 	uint16_t dlen;
@@ -878,45 +937,38 @@ uint16_t dec_indication(const uint8_t *pdu, int len, uint16_t *handle,
 	dlen = MIN(len - min_len, vlen);
 
 	if (handle)
-		*handle = att_get_u16(&pdu[1]);
+		*handle = get_le16(&pdu[1]);
 
 	memcpy(value, &pdu[3], dlen);
 
 	return dlen;
 }
 
-uint16_t enc_confirmation(uint8_t *pdu, int len)
+uint16_t enc_confirmation(uint8_t *pdu, size_t len)
 {
-	const uint16_t min_len = sizeof(pdu[0]);
-
 	if (pdu == NULL)
 		return 0;
 
-	if (len < min_len)
-		return 0;
-
+	/* Attribute Opcode (1 octet) */
 	pdu[0] = ATT_OP_HANDLE_CNF;
 
-	return min_len;
+	return 1;
 }
 
-uint16_t enc_mtu_req(uint16_t mtu, uint8_t *pdu, int len)
+uint16_t enc_mtu_req(uint16_t mtu, uint8_t *pdu, size_t len)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(mtu);
-
 	if (pdu == NULL)
 		return 0;
 
-	if (len < min_len)
-		return 0;
-
+	/* Attribute Opcode (1 octet) */
 	pdu[0] = ATT_OP_MTU_REQ;
-	att_put_u16(mtu, &pdu[1]);
+	/* Client Rx MTU (2 octets) */
+	put_le16(mtu, &pdu[1]);
 
-	return min_len;
+	return 3;
 }
 
-uint16_t dec_mtu_req(const uint8_t *pdu, int len, uint16_t *mtu)
+uint16_t dec_mtu_req(const uint8_t *pdu, size_t len, uint16_t *mtu)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*mtu);
 
@@ -932,28 +984,25 @@ uint16_t dec_mtu_req(const uint8_t *pdu, int len, uint16_t *mtu)
 	if (pdu[0] != ATT_OP_MTU_REQ)
 		return 0;
 
-	*mtu = att_get_u16(&pdu[1]);
+	*mtu = get_le16(&pdu[1]);
 
 	return min_len;
 }
 
-uint16_t enc_mtu_resp(uint16_t mtu, uint8_t *pdu, int len)
+uint16_t enc_mtu_resp(uint16_t mtu, uint8_t *pdu, size_t len)
 {
-	const uint16_t min_len = sizeof(pdu[0]) + sizeof(mtu);
-
 	if (pdu == NULL)
 		return 0;
 
-	if (len < min_len)
-		return 0;
-
+	/* Attribute Opcode (1 octet) */
 	pdu[0] = ATT_OP_MTU_RESP;
-	att_put_u16(mtu, &pdu[1]);
+	/* Server Rx MTU (2 octets) */
+	put_le16(mtu, &pdu[1]);
 
-	return min_len;
+	return 3;
 }
 
-uint16_t dec_mtu_resp(const uint8_t *pdu, int len, uint16_t *mtu)
+uint16_t dec_mtu_resp(const uint8_t *pdu, size_t len, uint16_t *mtu)
 {
 	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*mtu);
 
@@ -969,7 +1018,176 @@ uint16_t dec_mtu_resp(const uint8_t *pdu, int len, uint16_t *mtu)
 	if (pdu[0] != ATT_OP_MTU_RESP)
 		return 0;
 
-	*mtu = att_get_u16(&pdu[1]);
+	*mtu = get_le16(&pdu[1]);
 
 	return min_len;
 }
+
+uint16_t enc_prep_write_req(uint16_t handle, uint16_t offset,
+					const uint8_t *value, size_t vlen,
+					uint8_t *pdu, size_t len)
+{
+	const uint16_t min_len = sizeof(pdu[0]) + sizeof(handle) +
+								sizeof(offset);
+
+	if (pdu == NULL)
+		return 0;
+
+	if (vlen > len - min_len)
+		vlen = len - min_len;
+
+	pdu[0] = ATT_OP_PREP_WRITE_REQ;
+	put_le16(handle, &pdu[1]);
+	put_le16(offset, &pdu[3]);
+
+	if (vlen > 0) {
+		memcpy(&pdu[5], value, vlen);
+		return min_len + vlen;
+	}
+
+	return min_len;
+}
+
+uint16_t dec_prep_write_req(const uint8_t *pdu, size_t len, uint16_t *handle,
+				uint16_t *offset, uint8_t *value, size_t *vlen)
+{
+	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*handle) +
+							sizeof(*offset);
+
+	if (pdu == NULL)
+		return 0;
+
+	if (handle == NULL || offset == NULL || value == NULL || vlen == NULL)
+		return 0;
+
+	if (len < min_len)
+		return 0;
+
+	if (pdu[0] != ATT_OP_PREP_WRITE_REQ)
+		return 0;
+
+	*handle = get_le16(&pdu[1]);
+	*offset = get_le16(&pdu[3]);
+
+	*vlen = len - min_len;
+	if (*vlen > 0)
+		memcpy(value, pdu + min_len, *vlen);
+
+	return len;
+}
+
+uint16_t enc_prep_write_resp(uint16_t handle, uint16_t offset,
+					const uint8_t *value, size_t vlen,
+					uint8_t *pdu, size_t len)
+{
+	const uint16_t min_len = sizeof(pdu[0]) + sizeof(handle) +
+								sizeof(offset);
+
+	if (pdu == NULL)
+		return 0;
+
+	if (vlen > len - min_len)
+		vlen = len - min_len;
+
+	pdu[0] = ATT_OP_PREP_WRITE_RESP;
+	put_le16(handle, &pdu[1]);
+	put_le16(offset, &pdu[3]);
+
+	if (vlen > 0) {
+		memcpy(&pdu[5], value, vlen);
+		return min_len + vlen;
+	}
+
+	return min_len;
+}
+
+uint16_t dec_prep_write_resp(const uint8_t *pdu, size_t len, uint16_t *handle,
+				uint16_t *offset, uint8_t *value, size_t *vlen)
+{
+	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*handle) +
+								sizeof(*offset);
+
+	if (pdu == NULL)
+		return 0;
+
+	if (handle == NULL || offset == NULL || value == NULL || vlen == NULL)
+		return 0;
+
+	if (len < min_len)
+		return 0;
+
+	if (pdu[0] != ATT_OP_PREP_WRITE_REQ)
+		return 0;
+
+	*handle = get_le16(&pdu[1]);
+	*offset = get_le16(&pdu[3]);
+	*vlen = len - min_len;
+	if (*vlen > 0)
+		memcpy(value, pdu + min_len, *vlen);
+
+	return len;
+}
+
+uint16_t enc_exec_write_req(uint8_t flags, uint8_t *pdu, size_t len)
+{
+	if (pdu == NULL)
+		return 0;
+
+	if (flags > 1)
+		return 0;
+
+	/* Attribute Opcode (1 octet) */
+	pdu[0] = ATT_OP_EXEC_WRITE_REQ;
+	/* Flags (1 octet) */
+	pdu[1] = flags;
+
+	return 2;
+}
+
+uint16_t dec_exec_write_req(const uint8_t *pdu, size_t len, uint8_t *flags)
+{
+	const uint16_t min_len = sizeof(pdu[0]) + sizeof(*flags);
+
+	if (pdu == NULL)
+		return 0;
+
+	if (flags == NULL)
+		return 0;
+
+	if (len < min_len)
+		return 0;
+
+	if (pdu[0] != ATT_OP_EXEC_WRITE_REQ)
+		return 0;
+
+	*flags = pdu[1];
+
+	return min_len;
+}
+
+uint16_t enc_exec_write_resp(uint8_t *pdu)
+{
+	if (pdu == NULL)
+		return 0;
+
+	/* Attribute Opcode (1 octet) */
+	pdu[0] = ATT_OP_EXEC_WRITE_RESP;
+
+	return 1;
+}
+
+uint16_t dec_exec_write_resp(const uint8_t *pdu, size_t len)
+{
+	const uint16_t min_len = sizeof(pdu[0]);
+
+	if (pdu == NULL)
+		return 0;
+
+	if (len < min_len)
+		return 0;
+
+	if (pdu[0] != ATT_OP_EXEC_WRITE_RESP)
+		return 0;
+
+	return len;
+}
